package com.shilohmorris.dctm.plugin;

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ModalityState;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.fileTypes.FileTypeManager;
import com.intellij.openapi.fileTypes.StdFileTypes;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.ProjectRootManager;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.VirtualFileAdapter;
import com.intellij.openapi.vfs.VirtualFileEvent;
import com.intellij.openapi.vfs.VirtualFileManager;
import com.intellij.openapi.wm.ToolWindow;
import com.intellij.openapi.wm.ToolWindowAnchor;
import com.intellij.openapi.wm.ToolWindowManager;
import com.intellij.util.ui.UIUtil;
import com.shilohmorris.dctm.plugin.model.ComponentModel;
import com.shilohmorris.dctm.plugin.model.ComponentModelList;
import com.shilohmorris.dctm.plugin.ui.ComponentTreePanel;
import com.shilohmorris.dctm.plugin.ui.TreeNodeItem;
import com.shilohmorris.dctm.plugin.utils.PluginUtils;
import com.shilohmorris.dctm.plugin.utils.VirtualFileHandler;
import com.shilohmorris.dctm.plugin.utils.OpenNewEditorRunnable;
import org.apache.log4j.Logger;
import javax.swing.*;
import javax.swing.tree.DefaultMutableTreeNode;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;


/**
 * ****************************************************************************
 * Copyright 2006 Shiloh Morris                                               *
 *                                                                            *
 * Licensed under the Apache License, Version 2.0 (the "License");            *
 * you may not use this file except in compliance with the License.           *
 * You may obtain a copy of the License at                                    *
 *                                                                            *
 *     http://www.apache.org/licenses/LICENSE-2.0                             *
 *                                                                            *
 * Unless required by applicable law or agreed to in writing, software        *
 * distributed under the License is distributed on an "AS IS" BASIS,          *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   *
 * See the License for the specific language governing permissions and        *
 * limitations under the License.                                             *
 * ****************************************************************************
 */

/**
 * WdkPlugin General Information<br/><br/>
 *
 * <p>The Dctm Webtop component viewer plugin allows
 * the user to view webtop component parts, whether
 * they be java files, jsp files, or config files, in
 * a central convenient view.  No more diving through
 * folder after folder looking for the pieces that
 * you need.</p>
 *
 * <p>The plugin works by first determining the location
 * of key webtop folders, including "custom" and "config",
 * and then iterating and parsing each of the config files.</p>
 *
 * <p>With the info in the config files, we can build models of
 * each component in memory and to display them in the tree view.</p>
 *
 * <p>Do to corporate constraints, this project is JDK 1.4 compliant</p>
 *
 */
public class WdkPlugin implements ProjectComponent {

    // ------------------------------ FIELDS ------------------------------
    /**
     * ID of the plugin tool window
     */
    public static final String TOOL_WINDOW_ID = "WDK Components";

    /**
     * Name of the Webtop custom directory
     */
    public static final String CUSTOM_FOLDER_NAME = "custom";

    /**
     * Name of the config directory within the custom directory
     */
    public static final String CONFIG_FOLDER_NAME = "config";

    /**
     * Name of the configuration folder
     */
    public static final String STRINGS_FOLDER_NAME = "strings";

    /**
     * Reference to the current IDEA Project
     */
    private Project currentProject;

    /**
     * The customized JPanel that contains the plugin functionality
     */
    private ComponentTreePanel toolPanel = new ComponentTreePanel(new MouseEventHandler());

    /**
     * Convenience class to hold each component reference
     */
    private ComponentModelList componentModels = new ComponentModelList();

    /**
     * This is a map of Virtual File name -> component names
     * used when a configuration file is deleted.  Since the
     * contents are gone before we get notice, we must use
     * this map file to know what to delete from the tree view.
     */
    private Map fileToComponentMap = new HashMap();

    /**
     * Reference to the Webtop custom directory
     */
    private VirtualFile customDirectory;

    /**
     * Reference to the Webtop Strings directory
     */
    private VirtualFile stringsDirectory;

    /**
     * Reference to the Custom configuration directory
     */
    private VirtualFile customConfigDirectory;

    /**
     * Active logging instance
     */
    private Logger log = Logger.getLogger(this.getClass());

    /**
     * Flag which indicates whether the plugin is loaded or not
     */
    private boolean loaded;

    /**
     * Text which appears in tool window title
     */
    private String toolWindowTitle = "Tree View";


// --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Creates an instance of the WdkPlugin.  Called by the lugin framework
     * of IDEA
     * @param project current IDEA Project
     */
    public WdkPlugin(Project project) {
        currentProject = project;
    }

// ------------------------ INTERFACE METHODS ------------------------

// --------------------- Interface BaseComponent ---------------------


    /**
     * Returns the component name to the plugin framework
     * @return plugin's name
     */
    public String getComponentName() {
        return "WebtopComponentViewer.viewer";
    }

    public void initComponent() {
        // do nothing
    }

    public void disposeComponent() {
        // do nothing
    }

// --------------------- Interface ProjectComponent ---------------------

    /**
     * <p>Event handler for project opened</p>
     *
     * <p>If the plugin fails to find the custom directory
     * then the plugin will not initialize and the file
     * handlers will not attached.</p>
     */
    public void projectOpened() {
        loaded = populateBaseDirectories();
        if(loaded) {
            initToolWindow();
        }
    }

    /**
     * Event handler for project closed
     */
    public void projectClosed() {
        if(loaded) {
            unregisterToolWindow();
        }
    }

// -------------------------- OTHER METHODS --------------------------

    /**
     * Initialize GUI elements and register them with IDEA
     */
    private void initToolWindow() {

        VfsListener fileChangeListener = new VfsListener();
        VirtualFileManager.getInstance().addVirtualFileListener(fileChangeListener);
        ToolWindowManager toolWindowManager = ToolWindowManager.getInstance(currentProject);
        toolPanel.setBackground(UIUtil.getTreeTextBackground());
        ToolWindow toolWindow = toolWindowManager.registerToolWindow(TOOL_WINDOW_ID, toolPanel, ToolWindowAnchor.LEFT);

        toolWindow.setTitle(toolWindowTitle);
    }

    /**
     * Initializes the member variables associated with important
     * WebTop paths.
     *
     * @return whether or not the base directories were found
     */
    private boolean populateBaseDirectories() {

        boolean baseDirectoriesFound;
        ProjectRootManager projectRootManager = ProjectRootManager.getInstance(currentProject);
        VirtualFile[] sourceRoots = projectRootManager.getContentRoots();

        // Create a handler to find the custom directory
        VirtualFileHandler handler = new CustomDirHandler();
        List handlersList = new ArrayList();
        handlersList.add(handler);

        // iterate through the source roots looking for the custom dir
        // the handler will assign a value if it stumbles across it
        for (int i = 0; i < sourceRoots.length; i++) {
            VirtualFile sourceRoot = sourceRoots[i];
            try {
                iterateChildren(sourceRoot, handlersList);
            } catch (Exception e) {
                handleError(e, "Error handling source path: " + sourceRoot.getName());
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("Finished");
            log.debug("Custom: " + customDirectory);
            log.debug("Config: " + customConfigDirectory);
            log.debug("String: " + stringsDirectory);
        }

        if (customDirectory != null) {
            baseDirectoriesFound = true;
            handlersList = new ArrayList();
            handlersList.add(new ConfigFileHandler());

            try {
                iterateChildren(customDirectory, handlersList);
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            baseDirectoriesFound = false;
        }
        return baseDirectoriesFound;
    }

    /**
     * <p>Conditioanlly iterates through the virtual file and children
     * executing the handlers.  If any handler returns a TRUE value
     * the execution of the iteration ceases.</p>
     *
     * <p>TODO:  The multiple returns in the method is a bad-thing(tm) -- fix</p>
     *
     * @param virtualFile starting point from which to start
     * @param handlers    List of handlers to execute against each VirtualFile
     * @throws Exception failed to locate or read a file
     */
    private void iterateChildren(VirtualFile virtualFile, List handlers) throws Exception {
        for (Iterator handlerIt = handlers.iterator(); handlerIt.hasNext();) {
            VirtualFileHandler currentHandler = (VirtualFileHandler) handlerIt.next();
            boolean foundIt = currentHandler.handle(virtualFile);
            if (foundIt) return;
        }
        VirtualFile[] children = virtualFile.getChildren();
        if (children == null) return;
        for (int i = 0; i < children.length; i++) {
            VirtualFile child = children[i];
            for (Iterator handlerIt = handlers.iterator(); handlerIt.hasNext();) {
                VirtualFileHandler currentHandler = (VirtualFileHandler) handlerIt.next();
                boolean foundIt = currentHandler.handle(child);
                if (foundIt) return;
            }
            if (child.isDirectory()) {
                iterateChildren(child, handlers);
            }
        }
    }

    /**
     * Returns the specified file's contents as a String
     *
     * @param file to extract the contents from
     * @return string contents of file
     * @throws IOException error reading file's contents
     */
    private String getFileContents(VirtualFile file) throws IOException {
        BufferedReader reader = new BufferedReader(new InputStreamReader(file.getInputStream()));
        String line;
        StringBuffer contents = new StringBuffer();
        while ((line = reader.readLine()) != null) {
            contents.append(line);
        }
        return contents.toString();
    }

    /**
     * Provide a single facility to handle error messages that occur during
     * initiailization and use of the component.
     * @param e exception being thrown
     * @param message User-Friendly error message
     */
    private void handleError(Exception e, String message) {
        log.error(e);
        toolPanel.displayError(message);
    }

    /**
     * Removes the Tool Window items
     */
    private void unregisterToolWindow() {
        ToolWindowManager toolWindowManager = ToolWindowManager.getInstance(currentProject);
        toolWindowManager.unregisterToolWindow(TOOL_WINDOW_ID);
    }

    // -------------------------- INNER CLASSES --------------------------

    /**
     * Handles new file and deleted file events for the component
     */
    private class VfsListener extends VirtualFileAdapter {
        public void fileCreated(VirtualFileEvent event) {
            try {
                ConfigFileHandler handler = new ConfigFileHandler();
                handler.handle(event.getFile());

            } catch (Exception e) {
                handleError(e, "Error handling new File: " + event.getFileName());
            }
        }

        /**
         * If the file is a config file, delete the current tree
         * entry and re-read/add it back.
         * @param event file changed
         */
        public void contentsChanged(VirtualFileEvent event) {
            VirtualFile file = event.getFile();
            FileTypeManager fileTypeManager = FileTypeManager.getInstance();
            VirtualFile parentFile = file.getParent();
            if (!fileTypeManager.isFileIgnored(file.getName())
                    && fileTypeManager.getFileTypeByFile(file) == StdFileTypes.XML
                    && parentFile != null
                    && parentFile.getName().equalsIgnoreCase(CONFIG_FOLDER_NAME)) {
                try {
                    FileDeletionHandler delHandler = new FileDeletionHandler();
                    delHandler.handle(event.getFile());
                    ConfigFileHandler handler = new ConfigFileHandler();
                    handler.handle(event.getFile());

                } catch (Exception e) {
                    handleError(e, "Error handling new File: " + event.getFileName());
                }
            }
        }

        /**
         * Handle the notification that a file has been deleted
         * @param event
         */
        public void fileDeleted(VirtualFileEvent event) {
            try {
                FileDeletionHandler handler = new FileDeletionHandler();
                handler.handle(event.getFile());
            } catch (Exception e) {
                handleError(e, "Error handling deleted File: " + event.getFileName());
            }
        }
    }

    /**
     * Handler class that looks at the virtual file
     * and determines whether or not to assign it to
     * the custom, strings, or config directories.
     */
    private class CustomDirHandler implements VirtualFileHandler {

        public boolean handle(VirtualFile directory) {
            boolean found = false;
            if (directory != null) {
                if (directory.getName().equalsIgnoreCase(CUSTOM_FOLDER_NAME)) {
                    found = true;
                    customDirectory = directory;
                    VirtualFile[] files = customDirectory.getChildren();
                    for (int i = 0; i < files.length; i++) {
                        VirtualFile currentFile = files[i];
                        if (currentFile.getName().equalsIgnoreCase(STRINGS_FOLDER_NAME)) {
                            stringsDirectory = currentFile;
                        } else if (currentFile.getName().equalsIgnoreCase(CONFIG_FOLDER_NAME)) {
                            customConfigDirectory = currentFile;
                        }
                    }

                }
            }
            return found;
        }
    }

    /**
     * Handler class for WebTop Component configuration classes.
     */
    private class ConfigFileHandler implements VirtualFileHandler {

        /**
         * Determines whether current virtual file is a webtop configuration
         * file.  If so it creates a new ComponentModel class and adds it
         * to the existing list of component classes.
         * @param file Virtual File to be checked
         * @return whether or not to continue looking
         */
        public boolean handle(VirtualFile file) {
            FileTypeManager fileTypeManager = FileTypeManager.getInstance();
            if (!fileTypeManager.isFileIgnored(file.getName())
                    && fileTypeManager.getFileTypeByFile(file) == StdFileTypes.XML) {

                try {
                    ComponentModel model = new ComponentModel(getFileContents(file), file.getName());

                    if (model.getComponentName() != null
                            && model.getComponentName().equalsIgnoreCase("null") == false) {
                        componentModels.add(model);
                        toolPanel.addNewComponent(model);
                        // Note:  The full path would be more exclusive, but in the VfsListener
                        //        Class the VfsFile.getPath() only returns the name
                        fileToComponentMap.put(file.getName(), model);
                    }
                } catch (Exception e) {
                    log.error(e);
                }
            }
            return false;
        }
    }

    /**
     * Handler class to clean up the tree when config deletions occur
     * TODO: Critical bug - the file is GONE when this file is called.  Need
     *                      to keep a map of File->component name and manage finding it that way
     */
    private class FileDeletionHandler implements VirtualFileHandler {

        public boolean handle(VirtualFile file) {
            FileTypeManager fileTypeManager = FileTypeManager.getInstance();
            if (!fileTypeManager.isFileIgnored(file.getName())
                    && fileTypeManager.getFileTypeByFile(file) == StdFileTypes.XML) {
                try {

                    ComponentModel model = (ComponentModel) fileToComponentMap.get(file.getName());
                    if(model != null) {
                        toolPanel.removeComponentModel(model);
                        fileToComponentMap.remove(model);
                    }

                } catch (Exception e) {
                    log.error(e);
                }
            } // end if potential config file
            return false;
        }
    }



    /**
     *  Handles mouse events fired from the TREE.
     *  TODO:  refactor the mouseClicked method to be less lengthy - handlers?
     */
    class MouseEventHandler extends MouseAdapter {

        public void mouseClicked(MouseEvent mouseEvent) {

            JTree tree = toolPanel.getTree();
            DefaultMutableTreeNode node = (DefaultMutableTreeNode)
                    tree.getLastSelectedPathComponent();

            if (mouseEvent.getClickCount() == 2 && node != null) {

                // open the selected file
                TreeNodeItem item = (TreeNodeItem) node.getUserObject();
                String fileName = item.getFileName();
                String fileType = item.getType();

                if( fileName != null && fileName.equalsIgnoreCase("null") == false)

                // ****  Handle Java file selection
                if (fileType.equals(TreeNodeItem.JAVA_FILE)) {
                    VirtualFile vFile = PluginUtils.fqnToSourceFile(currentProject, fileName);
                    if (vFile == null) {
                        if(log.isDebugEnabled()) log.debug("No file could be found for class" + fileName);
                    } else {
                        OpenNewEditorRunnable runner = new OpenNewEditorRunnable(vFile, currentProject);
                        ApplicationManager.getApplication().invokeLater(runner, ModalityState.NON_MMODAL);
                    }
                // ****  Handle config file selection
                } else if (fileType.equals(TreeNodeItem.CONFIG_FILE)) {
                    VirtualFile[] configs = customConfigDirectory.getChildren();
                    for (int i = 0; i < configs.length; i++) {
                        if (configs[i].getName().equalsIgnoreCase(fileName)) {
                            OpenNewEditorRunnable runner = new OpenNewEditorRunnable(configs[i], currentProject);
                            ApplicationManager.getApplication().invokeLater(runner, ModalityState.NON_MMODAL);
                        }
                    }
                // ****  Handle JSP file selection
                } else if (fileType.equals(TreeNodeItem.JSP_FILE)) {
                    VirtualFile file = PluginUtils.findFromPath(fileName, currentProject);
                    if (file != null) {
                        OpenNewEditorRunnable runner = new OpenNewEditorRunnable(file, currentProject);
                        ApplicationManager.getApplication().invokeLater(runner, ModalityState.NON_MMODAL);
                    } else {
                       if(log.isDebugEnabled()) log.debug("Cannot open Null JSP file: "  + fileName);
                    }
                // ****  Handle Strings file selection
                } else if (fileType.equals(TreeNodeItem.NLSPROPS)) {
                    VirtualFile file = PluginUtils.findFromPath(fileName, currentProject);
                    if (file != null) {
                        OpenNewEditorRunnable runner = new OpenNewEditorRunnable(file, currentProject);
                        ApplicationManager.getApplication().invokeLater(runner, ModalityState.NON_MMODAL);
                    } else {
                        if(log.isDebugEnabled()) log.debug("Cannot open Null NLSPRops file: " + fileName);
                    }
                }
            }
        }
    }
}
