package com.shilohmorris.dctm.plugin.ui;

import com.shilohmorris.dctm.plugin.model.ComponentModel;
import com.shilohmorris.dctm.plugin.model.JspPage;
import com.shilohmorris.dctm.plugin.utils.PluginUtils;

import javax.swing.*;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeSelectionModel;
import javax.swing.tree.TreeModel;
import javax.swing.tree.DefaultTreeModel;

import java.util.*;
import java.util.List;
import java.awt.*;
import java.awt.event.MouseListener;

/*
 *
 * ****************************************************************************
 * Copyright 2006 Shiloh Morris                                               *
 *                                                                            *
 * Licensed under the Apache License, Version 2.0 (the "License");            *
 * you may not use this file except in compliance with the License.           *
 * You may obtain a copy of the License at                                    *
 *                                                                            *
 *     http://www.apache.org/licenses/LICENSE-2.0                             *
 *                                                                            *
 * Unless required by applicable law or agreed to in writing, software        *
 * distributed under the License is distributed on an "AS IS" BASIS,          *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   *
 * See the License for the specific language governing permissions and        *
 * limitations under the License.                                             *
 * ****************************************************************************
 */

/**
 * <p>This is the primary ui component of the Component Viewer plugin.
 * It provides access to the JTree widget.</p>
 *
 * <p>The ComponentTreePanel is also configured to provide an error
 * message should a problem occur during parsing.</p>
 */
public class ComponentTreePanel extends JPanel {
// ------------------------------ FIELDS ------------------------------

    // Titles and/or headers for the tree list items
    private static final String ROOT_NODE_TITLE = "Custom WebTop Components";
    private static final String  COMPONENT_SOURCE_TITLE = "Component Source - ";
    private static final String CONFIGURATION_FILE_TITLE ="Configuration File - ";
    private static final String NLS_FILE_TITLE =  "NLS Properties - ";
    private static final String JSP_FILE_TITLE ="JSP Files";
    private static final String JSP_FILTER_TITLE = "Filter";

    // displays the components based on component model
    JTree tree;

    // root tree node
    DefaultMutableTreeNode root = new DefaultMutableTreeNode(
            new TreeNodeItem(
                ROOT_NODE_TITLE, "", TreeNodeItem.ROOT
            )
    );

// --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Creates a new Component Tree Panel based on the passed in list of
     * ComponentModel objects.
     *
     */
    public ComponentTreePanel(MouseListener mouseListener) {
        super(new GridLayout(1,0));

        //setUpTreeElements(components);
        tree = new JTree(root);
        tree.setCellRenderer(new TreeRenderer());

        tree.getSelectionModel().setSelectionMode
                (TreeSelectionModel.SINGLE_TREE_SELECTION);
        tree.addMouseListener(mouseListener);

        JEditorPane editorPane = new JEditorPane();
        editorPane.setEditable(false);
        JScrollPane treeView = new JScrollPane(tree);

        // There will always be at least the root node.
        tree.expandRow(0);
        add(treeView);
    }

// -------------------------- OTHER METHODS --------------------------

    /**
     * Displays an error message in place of the normal
     * component tree.
     * @param message to be displayed
     */
    public void displayError(String message) {
        removeAll();
        setLayout(new BorderLayout());
        add(new ErrorPanel(message), BorderLayout.CENTER);
    }

    public JTree getTree() {
        return tree;
    }

    public void setTree(JTree tree) {
        this.tree = tree;
    }

    /**
     * Completely re-created the tree widget and re-applies all components currently
     * residing in the ComponentModel list
     * @param componentModels to create the tree nodes from
     */
    private void refreshComponentTree(List componentModels) {
        root.removeAllChildren();
        for(Iterator comps = componentModels.iterator(); comps.hasNext();) {
            ComponentModel model = (ComponentModel) comps.next();
            this.addNewComponent(model);
        }
    }

    /**
     * Adds a new component node to the existing tree model root.
     * Once the component is added, the tree is refreshed.
     *
     * TODO:  Break up this method - too long - factory pattern?
     *
     * @param model to be added
     */
    public void addNewComponent(ComponentModel model) {
        if( model != null) {
                DefaultMutableTreeNode compRoot = new DefaultMutableTreeNode(
                    new TreeNodeItem(
                        model.getComponentName(), "", TreeNodeItem.COMP_ROOT
                    )
                );

                if( model.getClassName() != null || model.getClassName().equalsIgnoreCase("null") == false) {
                    DefaultMutableTreeNode clazzNode  = new DefaultMutableTreeNode(
                        new TreeNodeItem(
                            COMPONENT_SOURCE_TITLE + model.getClassName(), model.getClassName(), TreeNodeItem.JAVA_FILE
                        ));
                    compRoot.add(clazzNode);
                }

                DefaultMutableTreeNode configNode  = new DefaultMutableTreeNode(
                    new TreeNodeItem(
                        CONFIGURATION_FILE_TITLE + model.getConfigFileName(), model.getConfigFileName(),
                            TreeNodeItem.CONFIG_FILE
                    ));
                compRoot.add(configNode);

               if( model.getStringsFile() != null && model.getStringsFile().length() > 0) {
                   DefaultMutableTreeNode stringNode = new DefaultMutableTreeNode(
                        new TreeNodeItem(
                            NLS_FILE_TITLE + model.getStringsFile(),
                                PluginUtils.translateToUrl(model.getStringsFile()) + ".properties",
                                TreeNodeItem.NLSPROPS
                        )
                    );
                   compRoot.add(stringNode);
                }



                // JSP nodes are slightly different in the sense that multiple
                // files can exist and they can be divided into "filter groups".
                //  The code below creates the ROOT->Component->JSP FILES->*FILTER->*FILES
                //  structure.
                DefaultMutableTreeNode jspFilesNode = new DefaultMutableTreeNode(
                    new TreeNodeItem(
                        JSP_FILE_TITLE, "", TreeNodeItem.JSP_FOLDER
                    )
                );

                Map filterNodesMap = new TreeMap();
                for(Iterator it = model.getJspFileList().iterator(); it.hasNext();) {
                    JspPage page = (JspPage) it.next();
                    String filter = page.getFilter();

                    // show the page name followed by the path stripped file name
                    String nodeDisplay = page.getName() + " - "
                            + page.getFilename().substring(page.getFilename().lastIndexOf("/") + 1);

                    DefaultMutableTreeNode jspNode = new DefaultMutableTreeNode(
                        new TreeNodeItem(
                            nodeDisplay, page.getFilename(), TreeNodeItem.JSP_FILE
                        )
                    );
                    DefaultMutableTreeNode filterNode = null;
                    if(filter != null ) {
                        if(filterNodesMap.containsKey(page.getFilter())) {
                            filterNode = (DefaultMutableTreeNode) filterNodesMap.get(filter);
                        } else {
                            filterNode = new DefaultMutableTreeNode(filter);
                            filterNode = new DefaultMutableTreeNode(
                                new TreeNodeItem(
                                    JSP_FILTER_TITLE, page.getFilter(), TreeNodeItem.JSP_FILTER
                                )
                            );
                            filterNodesMap.put(filter, filterNode);
                        }
                        filterNode.add(jspNode);
                    } else {
                        jspFilesNode.add(jspNode);
                    }
                    for(Iterator filterIt = filterNodesMap.values().iterator(); filterIt.hasNext();) {
                        DefaultMutableTreeNode currentFilterNode = (DefaultMutableTreeNode) filterIt.next();
                        jspFilesNode.add(currentFilterNode);
                    }
                }
                if(jspFilesNode.isLeaf() == false) {
                    compRoot.add(jspFilesNode);
                }

                root.add(compRoot);
        }
        refreshTreeModel();
    }

    /**
     * Removes the selected model from the tree's node structure.
     * Refreshes the tree model so that the specified component
     * is no longer visible.
     * @param model to be removed
     */
    public void removeComponentModel(ComponentModel model) {
        TreeModel treeModel = tree.getModel();
        DefaultMutableTreeNode rootNode = (DefaultMutableTreeNode) treeModel.getRoot();
        for(int i = 0; i < rootNode.getChildCount(); i++) {
            DefaultMutableTreeNode currentNode = (DefaultMutableTreeNode) rootNode.getChildAt(i);
            if(currentNode.toString().equalsIgnoreCase(model.getComponentName())) {
                rootNode.remove(currentNode);
                
            }
        }
        refreshTreeModel();

    }

    /**
     * Forces redrawing of the JTree with the existing JTree model.  To Add or Remove
     * components, you must use the appropriate method (they call this method by default).
     */
    private void refreshTreeModel() {
        DefaultTreeModel defaultTreeModel=(DefaultTreeModel)tree.getModel();
        defaultTreeModel.reload(root);
    }
}
